

class MotorLogs:
    def decode_motor_message(self, request):
        pass

    def decode_motor_response(self, response):
        if response:
            data = response.data
            print(f"Response ID: {response.arbitration_id}")
            print(f"Response Data: {[hex(b) for b in data]}")
            if data[0] == 0xA1:
                iq_current = data[1] | (data[2] << 8)
                iq_current = iq_current if iq_current < 32768 else iq_current - 65536  # convert to signed
                print(f"Torque Current Response: {iq_current} mA")
            else:
                print("Unexpected response")
        else:
            print("No response received")


class MotorInterface:
    def init(self):
        raise NotImplementedError

    def get_PID(self):
        raise NotImplementedError

    def set_position(self, position):
        raise NotImplementedError

    def set_speed(self, speed):
        raise NotImplementedError



from NekoRMD.Control import Control
from NekoRMD.MotorInterface import MotorInterface
from NekoRMD.MotorType import MotorType
from NekoRMD.Telemetry import Telemetry
from NekoRMD.MotorLogs import MotorLogs
from NekoRMD.ican import I_CAN
import time


class NekoRMD(MotorInterface):
    def __init__(self, address, motor_type):
        if not isinstance(motor_type, MotorType):
            raise ValueError("Invalid motor type")
        
        self.io = I_CAN(address)
        self.motor_type = motor_type
        self.control = Control(self.motor_type, self.io, address)
        self.telemetry = Telemetry(self.motor_type, self.io, address)
        self.logs = MotorLogs()

    WAIT_AFTER_INIT = True

    def setup(self):
        return self.io.setup()


    def init(self):
        if self.motor_type == MotorType.RMD8x_pro:
            message = [0x88, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]
            response = self.io.send_cmd(message, 0.01)
        
            if not(response.arbitration_id == self.motor_address and response.data[0] == 0x88):
                return None
            
            print(f"Motor of type {self.motor_type.name} initialized at address {self.io.identifier}")

            if self.WAIT_AFTER_INIT:
                print("please, wait 10 sec...")
                time.sleep(10)

            return response

    def restart(self):
        return self.init()

    def wait(self, seconds):
        time.sleep(seconds)



import os
import time
import can


class I_CAN:
    def __init__(self, identifier):
        self.bus = None
        self.identifier = identifier
    
    def setup(self):
        try:
            os.system("sudo /sbin/ip link set can0 up type can bitrate 1000000")
            time.sleep(0.1)
        except Exception as e:
            print(e)

        try:
            bus = can.interface.Bus(bustype='socketcan', channel='can0')
            self.bus = bus
        except OSError:
            print('err: PiCAN board was not found.')
            exit()
        except Exception as e:
            print(e)

        return self.bus

    def send_cmd(self, data, delay):
        """
        Отправка команды через интерфейс CAN.

        Параметры:
        ----------
        msg : список
            Команда в формате списка байтов.
        delay : float
            Задержка перед отправкой следующей команды (по умолчанию 0.01 секунды).
        """
        
        if len(data) != 8:
            raise ValueError("Data length should be exactly 8 bytes")
        message = can.Message(arbitration_id=self.identifier,
                              data=data, is_extended_id=False)
        try:
            self.bus.send(message)
            time.sleep(delay)
            received_message = self.bus.recv()
            return received_message
        except can.CanError as e:
            print(f"Message NOT sent: {e}")


from NekoRMD.MotorType import MotorType
import struct
import time
from datetime import datetime

class Control:
    def __init__(self, motor_type, io, motor_address):
        if not isinstance(motor_type, MotorType):
            raise ValueError("Invalid motor type")
        
        self.motor_type = motor_type
        self.io = io
        self.motor_address = motor_address

        self.position = self.Position(self, self.io, self.motor_address)
        self.speed = self.Speed(self, self.io, self.motor_address)
        self.Torque = self.Torque(self, self.io, self.motor_address)

    def freeze(self):
        message = [0x81, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]
        response = self.io.send_cmd(message, 0.01)
        if not(response.arbitration_id == self.motor_address and response.data[0] == 0x81):
            return None
        return response
    
    def unfreeze(self):
        message = [0x77, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]
        response = self.io.send_cmd(message, 0.01)
        if not(response.arbitration_id == self.motor_address and response.data[0] == 0x77):
            return None
        return response

    class Position:
        def __init__(self, control, io, motor_address):
            self.control = control
            self.io = io
            self.motor_address = motor_address

        def __get_motor_position(self):
            message = [0x90, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]
            response = self.io.send_cmd(message, delay=0.01)

            if not(response.arbitration_id == self.motor_address and response.data[0] == 0x90):
                return None
            
            timestamp = response.timestamp
            data = response.data
            return timestamp, data

        def __parse_position_data(self, data):
            """
            Конвертирует данные ответа с двигателя в понятный формат
            :param data: список байт данных
            :return: текущая позиция энкодера, исходная позиция энкодера, нулевое смещение энкодера
            """
            current_position = data[2] << 8 | data[1]
            original_position = data[4] << 8 | data[3]
            zero_offset = data[6] << 8 | data[5]
            return current_position, original_position, zero_offset

        def get_position_data(self):
            timestamp, data = self.__get_motor_position()
            current_position, _, _ = self.__parse_position_data(data)
            return timestamp, current_position
        
        def __print_position_data(self):
            timestamp, data = self.__get_motor_position()
            current_position, original_position, zero_offset = self.__parse_position_data(data)
            
            timestamp_time = datetime.fromtimestamp(timestamp)
            current_time = datetime.now()
            time_difference = current_time - timestamp_time

            print("Current time:", current_time)
            print(f"Timestamp: {timestamp}")
            print("Delta time:", time_difference)
            print(f"Current Position: {current_position}")
            print(f"Original Position: {original_position}")
            print(f"Zero Offset: {zero_offset}")
            print("-" * 30)

        def print_position_data(self, is_loop=False):
            self.__print_position_data()
            if is_loop:
                while True:
                    self.__print_position_data()
                    now = datetime.now()

                    detailed_time = now.strftime("%Y-%m-%d %H:%M:%S.%f")
                    print("Текущее время с микросекундами:", detailed_time)
                    time.sleep(0.2)


        def set_increment_position(self, max_speed, angle):
            """
            Управление приращением позиции (многооборотный режим).

            Параметры:
            ----------
            max_speed : int
                Максимальная скорость в RPM.
            angle : int
                Приращение угла в сотых долях градуса.

            Пример:
            -------
            motor.control.position.set_increment_position(max_speed=1000, angle=3600)
            """
                
            msg = [0xA8, 
                   0x00, 
                   max_speed & 0xFF, 
                   (max_speed >> 8) & 0xFF, 
                   angle & 0xFF, 
                   (angle >> 8) & 0xFF, 
                   (angle >> 16) & 0xFF, 
                   (angle >> 24) & 0xFF]
            response = self.io.send_cmd(msg, 0.1)

            if not(response.arbitration_id == self.motor_address and response.data[0] == 0xA8):
                return None
            return response

        def set_single_turn_position(self, direction, max_speed, angle):
            """
            Управление положением на одном обороте.

            Параметры:
            ----------
            direction : int
                Направление вращения (0x00 - по часовой стрелке, 0x01 - против часовой стрелки).
            max_speed : int
                Максимальная скорость в RPM.
            angle : int
                Угол управления в сотых долях градуса.

            Пример:
            -------
            motor.control.position.set_single_turn_position(0x01, 1000, 3800)
            """
            msg = [0xA6, 
                   direction, 
                   max_speed & 0xFF, 
                   (max_speed >> 8) & 0xFF, 
                   angle & 0xFF, 
                   (angle >> 8) & 0xFF, 
                   0x00, 
                   0x00]
            response = self.io.send_cmd(msg, 0.1)
            if not(response.arbitration_id == self.motor_address and response.data[0] == 0xA6):
                return None
            return response

        def set_absolute_position(self, max_speed, angle):
            """
            Управление абсолютной позицией (многооборотный режим).

            Параметры:
            ----------
            max_speed : int
                Максимальная скорость в RPM.
            angle : int
                Абсолютный угол в сотых долях градуса.

            Пример:
            -------
            motor.control.position.set_absolute_position(max_speed=1000, angle=7200)
            """
                
            message = [0xA4, 
                       0x00, 
                       max_speed & 0xFF, 
                       (max_speed >> 8) & 0xFF, 
                       angle & 0xFF, 
                       (angle >> 8) & 0xFF, 
                       (angle >> 16) & 0xFF, 
                       (angle >> 24) & 0xFF]
            response = self.io.send_cmd(message, 0.1)
            if not(response.arbitration_id == self.motor_address and response.data[0] == 0xA4):
                return None
            return response


    class Speed:
        def __init__(self, control, io, motor_address):
            self.control = control
            self.io = io
            self.motor_address = motor_address
        
        def set_speed(self, speed=0):
            """
            Sends a speed control command to the motor via the CAN bus.

            This function forms and sends command 0xA2 to control the speed of the motor output shaft.
            The speed value is transmitted in the int32_t format and corresponds to a speed of 0.01 degrees/second per least significant bit (LSB).

            Parameters:
            - speed (float): The desired speed in degrees per second. The value can be positive or negative to control the direction of rotation.

            Notes:
            - The speed value will be rounded to the nearest integer.
            - A speed of 0 will stop the motor.
            - Check the range of allowable speeds for your motor in the technical documentation.

            Example usage:
            motor.control.speed.set_speed(100)
            """
            speedControl = int(speed / 0.01)
            message = [0xA2, 0x00, 0x00, 0x00, 
                speedControl & 0xFF, 
                (speedControl >> 8) & 0xFF, 
                (speedControl >> 16) & 0xFF, 
                (speedControl >> 24) & 0xFF]
            # print("Request: " + str([hex(byte) for byte in msg]))   
            response = self.io.send_cmd(message, 0.1)
            if not(response.arbitration_id == self.motor_address and response.data[0] == 0xA2):
                return None
            return response

    class Torque:
        def __init__(self, control, io, motor_address):
            self.control = control
            self.io = io
            self.motor_address = motor_address
        
        def test(self):
            message = [0xA1, 0x00, 0x00, 0x00, 0x64, 0x00, 0x00, 0x00]
            response = self.io.send_cmd(message, 0.01)
            if not(response.arbitration_id == self.motor_address and response.data[0] == 0xA1):
                return None
            return response

        def __calculate_crc(self, data):
            crc = 0xFFFF
            for pos in data:
                crc ^= pos
                for _ in range(8):
                    if (crc & 1) != 0:
                        crc >>= 1
                        crc ^= 0xA001
                    else:
                        crc >>= 1
            return crc
        
        def set_current(self, current):
            command = 0xA1
            current_mA = int(current * 100)

            message = [
                command,
                0x00, # NULL
                current_mA & 0xFF, # Torque Current Low Byte
                (current_mA >> 8) & 0xFF, # Torque Current High Byte
                0x00, # NULL
                0x00, # NULL
                0x00, # NULL
                0x00  # NULL
            ]

            # Calculate CRC
            crc = self.__calculate_crc(message)
            message.append(crc & 0xFF) # CRC Low Byte
            message.append((crc >> 8) & 0xFF) # CRC High Byte

            # Ensure the message is exactly 8 bytes for CAN
            if len(message) > 8:
                message = message[:8]
            elif len(message) < 8:
                message += [0x00] * (8 - len(message))

            print([hex(num) for num in message])  # Print hex values for debugging

            response = self.io.send_cmd(message, 0.01)
            if not(response.arbitration_id == self.motor_address and response.data[0] == 0xA1):
                return None
            return response


from NekoRMD.MotorType import MotorType
import time
import struct

class Telemetry:
    def __init__(self, motor_type, io, motor_address):
        if not isinstance(motor_type, MotorType):
            raise ValueError("Invalid motor type")
        
        self.motor_type = motor_type
        self.io = io
        self.motor_address = motor_address

    # ==================================================
    # |
    # |                      Telemetry
    # |
    # ==================================================

    def read_acceleration(self):
        """
        Чтение значения ускорения.

        :return: Значение ускорения.
        """
        message = [0x42, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]
        response = self.io.send_cmd(message, 0.1)
        if response.arbitration_id == self.motor_address and response.data[0] == 0x42:
            acceleration = struct.unpack('<H', response.data[2:4])[0]
            return acceleration
        return None
    
    def read_software_version(self):
        """
        Чтение версии программного обеспечения.

        :return: Дата версии ПО.
        """
        message = [0xB2, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]
        response = self.io.send_cmd(message, 0.1)
        if response.arbitration_id == self.motor_address and response.data[0] == 0xB2:
            version_date = struct.unpack('<H', response.data[4:6])[0]
            return version_date
        return None
    
    def read_motor_temperature_and_torque(self):
        """
        Чтение температуры мотора и текущего крутящего момента.

        :return: Словарь с температурой, крутящим моментом, скоростью и углом.
        """
        message = [0xA2, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]
        response = self.io.send_cmd(message, 0.1)
        if response.arbitration_id == self.motor_address and response.data[0] == 0xA2:
            temp, iq, speed, angle = struct.unpack('<BHHH', response.data[1:])
            return {'Temp': temp, 'Iq': iq, 'Speed': speed, 'Angle': angle}
        return None
    
    def read_system_errors(self):
        """
        Чтение состояния ошибки системы.

        :return: Словарь с ошибками системы, тормозом, напряжением и температурой.
        """
        message = [0x9A, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]
        response = self.io.send_cmd(message, 0.1)
        if response.arbitration_id == self.motor_address and response.data[0] == 0x9A:
            if len(response.data) < 4:
                raise ValueError("Received data is too short for motor power.")
            power = struct.unpack('<H', response.data[2:4])[0]
            return power
        return None

    def read_motor_power(self):
        """
        Чтение мощности мотора.

        :return: Значение мощности.
        """
        message = [0x71, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]
        response = self.io.send_cmd(message, 0.1)
        if response.arbitration_id == self.motor_address and response.data[0] == 0x71:
            power = struct.unpack('<H', response.data[2:4])[0]
            return power
        return None


    # ==================================================
    # |
    # |                       Encoder
    # |
    # ==================================================


    # TODO: не корректно производится подсчет encoder_position
    def get_absolute_encoder(self):
        if (self.motor_type == MotorType.RMD8x_pro):
            message = [0x60, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]
        response = self.io.send_cmd(message, 0.01)

        # RESPONSE.DATA:
        #   
        # [0x60, 0x00, ENCODER_POS_1, ENCODER_POS_2, ENCODER_POS_3, ENCODER_POS_4, 0x00, 0x00]
        #
        # Байт 0: Код команды (0x60)
        # Байт 1: Статус ошибки (Error Status)
        # Байт 2-3: Момент текущего крутящего момента (Current Torque) (2 байта)
        # Байт 4-5: Скорость мотора (Motor Speed) (2 байта)
        # Байт 6-7: Текущая позиция (Current Position) (2 байта)

        print("Request: " + str([hex(byte) for byte in response.data]))  

        if response.arbitration_id == self.motor_address and response.data[0] == 0x60:
            encoder_position = (response.data[4] | (response.data[5] << 8) | (response.data[6] << 16) | (response.data[7] << 24))
            return encoder_position
        else:
            print("Invalid response")
        
    
    def __print_absolute_encoder(self):
        print(f"Multi-turn Encoder Position: {self.get_absolute_encoder()}")
    
    def print_absolute_encoder(self, is_loop=False, delay=1):
        self.__print_absolute_encoder()
        if is_loop:
            while True:
                self.__print_absolute_encoder()
                time.sleep(delay)
                print("=" * 30)

    def get_relative_encoder(self):
        if (self.motor_type == MotorType.RMD8x_pro):
            message = [0x90, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]
        response = self.io.send_cmd(message, 0.01)

        # RESPONSE.DATA
        #
        # [0x90, 0x00, ENCODER_LOW, ENCODER_HIGH, ENCODER_RAW_LOW, ENCODER_RAW_HIGH, ENCODER_OFFSET_LOW, ENCODER_OFFSET_HIGH]
        #
        # ENCODER_LOW и ENCODER_HIGH - младший и старший байты текущего положения энкодера, 
        # ENCODER_RAW_LOW и ENCODER_RAW_HIGH - младший и старший байты исходного положения энкодера, 
        # ENCODER_OFFSET_LOW и ENCODER_OFFSET_HIGH - младший и старший байты смещения нуля энкодера.

        if response.arbitration_id == self.motor_address and response.data[0] == 0x90:
            encoder_position = response.data[2] | (response.data[3] << 8)
            encoder_raw_position = response.data[4] | (response.data[5] << 8)
            encoder_offset = response.data[6] | (response.data[7] << 8)
            return encoder_position, encoder_raw_position, encoder_offset
        else:
            print("Invalid response")

    def __print_relative_encoder(self):
        encoder_position, encoder_raw_position, encoder_offset = self.get_relative_encoder()

        print(f"Encoder Position: {encoder_position}")
        print(f"Encoder Raw Position: {encoder_raw_position}")
        print(f"Encoder Offset: {encoder_offset}")
    
    def print_relative_encoder(self, is_loop=False, delay=1):
        self.__print_relative_encoder()
        if is_loop:
            while True:
                self.__print_relative_encoder()
                time.sleep(delay)
                print("=" * 30)



    # ==================================================
    # |
    # |                       PID
    # |
    # ==================================================
    
    def read_pid_params(self):
        """
        Читает PID параметры мотора.

        Эта функция отправляет команду для получения текущих PID параметров мотора и 
        возвращает их в виде словаря. Параметры включают коэффициенты пропорционального (KP) и 
        интегрального (KI) регулирования для текущего контура, контура скорости и контура позиции.

        Returns:
            dict: Словарь с PID параметрами в формате:
                {
                    "current_kp": int,
                    "current_ki": int,
                    "speed_kp": int,
                    "speed_ki": int,
                    "position_kp": int,
                    "position_ki": int
                }
                Если ответ некорректен, возвращается None.
        """

        message = [0x30, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]
        response = self.io.send_cmd(message, 0.1)

        if response.arbitration_id == self.motor_address and response.data[0] == 0x30:
            _, _, current_kp, current_ki, speed_kp, speed_ki, position_kp, position_ki = struct.unpack("<BBBBBBBB", response.data)
            
            return {
                "current_kp": current_kp,
                "current_ki": current_ki,
                "speed_kp": speed_kp,
                "speed_ki": speed_ki,
                "position_kp": position_kp,
                "position_ki": position_ki
            }
        else:
            print("Error: Invalid response")
            return None
        
    def __print_pid_params(self):
        pid_params = self.read_pid_params()
        if pid_params:
            print(f"Current PID params:")
            print(f"  Текущий контур: KP = {pid_params['current_kp']}, KI = {pid_params['current_ki']}")
            print(f"  Контур скорости: KP = {pid_params['speed_kp']}, KI = {pid_params['speed_ki']}")
            print(f"  Контур позиции: KP = {pid_params['position_kp']}, KI = {pid_params['position_ki']}")

    def print_pid_params(self, is_loop=False, delay=1):
        """
        Выводит PID параметры мотора с возможностью повторения.

        Эта функция вызывает __print_pid_params для вывода текущих PID параметров мотора. 
        Если параметр is_loop установлен в True, функция будет повторять вывод параметров с заданной задержкой.

        Args:
            is_loop (bool): Флаг, указывающий, следует ли повторять вывод параметров. По умолчанию False.
            delay (int): Задержка между повторениями вывода параметров в секундах. По умолчанию 1 секунда.
        """
        self.__print_pid_params()
        if is_loop:
            while True:
                self.__print_pid_params()
                time.sleep(delay)
                print("=" * 30)
    
    def __convert_to_8bit(self, value, scale=1.0):
        """
        Конвертирует значение с плавающей запятой в 8-битное целое.

        Эта функция принимает значение с плавающей запятой, умножает его на заданный масштаб и 
        возвращает результат как 8-битное целое число.

        Args:
            value (float): Значение для конвертации.
            scale (float): Масштабный коэффициент. По умолчанию 1.0.

        Returns:
            int: 8-битное целое число.
        """
        return int(value * scale) & 0xFF
    
    def write_pid_params_ram(self, current_kp, current_ki, speed_kp, speed_ki, position_kp, position_ki):
        """
        Записывает PID параметры в RAM мотора.

        Эта функция отправляет команду для записи PID параметров в оперативную память (RAM) мотора.

        Args:
            current_kp (float): Коэффициент пропорционального регулирования текущего контура.
            current_ki (float): Коэффициент интегрального регулирования текущего контура.
            speed_kp (float): Коэффициент пропорционального регулирования контура скорости.
            speed_ki (float): Коэффициент интегрального регулирования контура скорости.
            position_kp (float): Коэффициент пропорционального регулирования контура позиции.
            position_ki (float): Коэффициент интегрального регулирования контура позиции.

        Returns:
            response: Ответ устройства. Если ответ некорректен, возвращается None.
        """
        message = [
            0x31, 0x00,
            self.convert_to_8bit(current_kp),
            self.convert_to_8bit(current_ki),
            self.convert_to_8bit(speed_kp),
            self.convert_to_8bit(speed_ki),
            self.convert_to_8bit(position_kp),
            self.convert_to_8bit(position_ki)
        ]
        
        response = self.io.send_cmd(message, 0.1)
        if response.arbitration_id == self.motor_address and response.data[0] == 0x31:
            return response
        else:
            print("Error: Invalid response!")
            return None
    
    def write_pid_params_rom(self, current_kp, current_ki, speed_kp, speed_ki, position_kp, position_ki):
        """
        Записывает PID параметры в ROM мотора.

        Эта функция отправляет команду для записи PID параметров в постоянную память (ROM) мотора.

        Args:
            current_kp (float): Коэффициент пропорционального регулирования текущего контура.
            current_ki (float): Коэффициент интегрального регулирования текущего контура.
            speed_kp (float): Коэффициент пропорционального регулирования контура скорости.
            speed_ki (float): Коэффициент интегрального регулирования контура скорости.
            position_kp (float): Коэффициент пропорционального регулирования контура позиции.
            position_ki (float): Коэффициент интегрального регулирования контура позиции.

        Returns:
            response: Ответ устройства. Если ответ некорректен, возвращается None.
        """
        message = [
            0x32, 0x00,
            self.convert_to_8bit(current_kp),
            self.convert_to_8bit(current_ki),
            self.convert_to_8bit(speed_kp),
            self.convert_to_8bit(speed_ki),
            self.convert_to_8bit(position_kp),
            self.convert_to_8bit(position_ki)
        ]

        response = self.io.send_cmd(message, 0.1)
        if response.arbitration_id == self.motor_address and response.data[0] == 0x32:
            return response
        else:
            print("Error: Invalid response!")
            return None



from setuptools import setup, find_packages

setup(
    name='NekoRMD',
    version='0.2.0',
    author='Artem-Darius Weber',
    author_email='neko-rmd@k-lab.su',
    description='NekoRMD is a Python library for controlling RMD series motors through a CAN interface. It provides a wide range of functions for motor control and monitoring.',
    long_description=open('README.md', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/NekoRMD/NekoRMD',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)
